@@ -53,8 +53,26 @@ def _assign_bin(timestamp, bins: list[datetime.datetime]) -> int:
     return len(bins) - 2  # Assign to last bin if at the end
 
 
-def _best_trial(study: optuna.Study) -> optuna.trial.FrozenTrial:
-    return study.best_trial
+def _best_trial(
+    study: optuna.Study, dt: datetime.datetime | None = None
+) -> optuna.trial.FrozenTrial:
+    if dt is None:
+        return study.best_trial
+    min_trial = None
+    for trial in study.trials:
+        if trial.value is None:
+            continue
+        dt_idx = datetime.datetime.fromisoformat(trial.user_attrs[_IDX_USR_ATTR_KEY])
+        if dt_idx == dt:
+            if min_trial is None:
+                min_trial = trial
+            elif min_trial.value is None:
+                min_trial = trial
+            elif min_trial.value > trial.value:
+                min_trial = trial
+    if min_trial is None:
+        min_trial = study.best_trial
+    return min_trial
     # best_brier = min(study.best_trials, key=lambda t: t.values[1])
     # return best_brier
 
@@ -78,6 +96,7 @@ class Trainer(Fit):
         allowed_models: set[str] | None = None,
         max_false_positive_reduction_steps: int | None = None,
         correlation_chunk_size: int | None = None,
+        insert_nulls: bool = False,
     ):
         tqdm.tqdm.pandas()
 
@@ -180,6 +199,7 @@ class Trainer(Fit):
         self._allowed_models = allowed_models
         self._max_false_positive_reduction_steps = max_false_positive_reduction_steps
         self._correlation_chunk_size = correlation_chunk_size
+        self._insert_nulls = insert_nulls
 
     def _provide_study(self, column: str) -> optuna.Study:
         storage_name = f"sqlite:///{self._folder}/{column}/{_STUDYDB_FILENAME}"
@@ -282,7 +302,9 @@ class Trainer(Fit):
                     # Perform common reductions
                     start_reducer = time.time()
                     reducer = CombinedReducer(
-                        self.embedding_cols, self._correlation_chunk_size
+                        self.embedding_cols,
+                        self._correlation_chunk_size,
+                        self._insert_nulls,
                     )
                     reducer.set_options(trial, x)
                     x_train = reducer.fit_transform(x_train, y=y_train)
@@ -477,7 +499,7 @@ class Trainer(Fit):
                 if found:
                     last_processed_dt = test_dt
                     _fit(
-                        _best_trial(study),
+                        _best_trial(study, dt=test_dt),
                         test_df.copy(),
                         test_series,
                         True,
@@ -496,7 +518,7 @@ class Trainer(Fit):
 
                 if test_idx < start_validation_index:
 
-                    def validate_objctive(
+                    def validate_objective(
                         trial: optuna.Trial,
                         idx: datetime.datetime,
                         series: pd.Series,
@@ -506,7 +528,7 @@ class Trainer(Fit):
 
                     study.optimize(
                         functools.partial(
-                            validate_objctive, idx=test_idx, series=test_series
+                            validate_objective, idx=test_idx, series=test_series
                         ),
                         n_trials=1,
                         timeout=None
@@ -517,7 +539,15 @@ class Trainer(Fit):
                     break
 
                 _fit(
-                    _best_trial(study),
+                    study.best_trial,
+                    test_df.copy(),
+                    test_series,
+                    False,
+                    test_idx,
+                    False,
+                )
+                _fit(
+                    _best_trial(study, dt=test_dt),
                     test_df.copy(),
                     test_series,
                     True,
@@ -537,15 +567,18 @@ class Trainer(Fit):
             #    height=600,
             # )
             for target_name in target_names:
-                fig = optuna.visualization.plot_param_importances(
-                    study, target=lambda t: t.values[0], target_name=target_name
-                )
-                fig.write_image(
-                    os.path.join(column_dir, f"{target_name}_frontier.png"),
-                    format="png",
-                    width=800,
-                    height=600,
-                )
+                try:
+                    fig = optuna.visualization.plot_param_importances(
+                        study, target=lambda t: t.values[0], target_name=target_name
+                    )
+                    fig.write_image(
+                        os.path.join(column_dir, f"{target_name}_frontier.png"),
+                        format="png",
+                        width=800,
+                        height=600,
+                    )
+                except RuntimeError as exc:
+                    print(str(exc))
 
         if isinstance(y, pd.Series):
             _fit_column(y)
@@ -615,7 +648,9 @@ class Trainer(Fit):
                 folder = os.path.join(column_path, date_str)
 
                 reducer = CombinedReducer(
-                    self.embedding_cols, self._correlation_chunk_size
+                    self.embedding_cols,
+                    self._correlation_chunk_size,
+                    self._insert_nulls,
                 )
                 reducer.load(folder)
 
@@ -674,9 +709,23 @@ class Trainer(Fit):
                 if not os.path.isdir(date_path):
                     continue
                 try:
+                    reducer = CombinedReducer(
+                        self.embedding_cols,
+                        self._correlation_chunk_size,
+                        self._insert_nulls,
+                    )
+                    reducer.load(date_path)
+
                     model = ModelRouter(None, None)
                     model.load(date_path)
-                    feature_importances[date_str] = model.feature_importances(df)
+
+                    selector = Selector(model)
+                    selector.load(date_path)
+
+                    x_pred = reducer.transform(df)
+                    x_pred = selector.transform(x_pred)
+
+                    feature_importances[date_str] = model.feature_importances(x_pred)
                 except FileNotFoundError as exc:
                     logging.warning(str(exc))
 
